import os
import websocket
import json
from tiingo.exceptions import MissingRequiredArgumentError


class TiingoWebsocketClient:
    """
    from tiingo import TiingoWebsocketClient

    def cb_fn(msg):

        # Example response
        # msg = {
        #   "service":"iex" # An identifier telling you this is IEX data.
        #   The value returned by this will correspond to the endpoint argument.
        #
        #   # Will always return "A" meaning new price quotes. There are also H type Heartbeat msgs used to keep the connection alive
        #   "messageType":"A" # A value telling you what kind of data packet this is from our IEX feed.
        #
        #   # see https://api.tiingo.com/documentation/websockets/iex > Response for more info
        #   "data":[] # an array containing trade information and a timestamp
        #
        # }

        print(msg)

    subscribe = {
            'eventName':'subscribe',
            'authorization':'API_KEY_GOES_HERE',
            #see https://api.tiingo.com/documentation/websockets/iex > Request for more info
            'eventData': {
                'thresholdLevel':5
          }
    }
    # notice how the object isn't needed after using it
    # any logic should be implemented in the callback function
    TiingoWebsocketClient(subscribe,endpoint="iex",on_msg_cb=cb_fn)
    while True:pass
    """

    def __init__(self, config=None, endpoint=None, on_msg_cb=None):
        self._base_url = "wss://api.tiingo.com"
        self.config = {} if config is None else config

        try:
            api_key = self.config["authorization"]
        except KeyError:
            api_key = os.environ.get("TIINGO_API_KEY")
            self.config.update({"authorization": api_key})

        self._api_key = api_key
        if not (api_key):
            raise RuntimeError(
                "Tiingo API Key not provided. Please provide"
                " via environment variable or config argument."
                "Notice that this config dict takes the API Key as authorization "
            )

        self.endpoint = endpoint
        if not (
            self.endpoint == "iex" or self.endpoint == "fx" or self.endpoint == "crypto"
        ):
            raise AttributeError("Endpoint must be defined as either (iex,fx,crypto) ")

        self.on_msg_cb = on_msg_cb
        if not self.on_msg_cb:
            raise MissingRequiredArgumentError(
                "please define on_msg_cb It's a callback that gets called when new messages arrive "
                "Example:"
                "def cb_fn(msg):"
                "    print(msg)"
            )

        websocket.enableTrace(False)

        ws = websocket.WebSocketApp(
            "{0}/{1}".format(self._base_url, self.endpoint),
            on_message=self.get_on_msg_cb(),
            on_error=self.on_error,
            on_close=self.on_close,
            on_open=self.get_on_open(self.config),
        )
        ws.run_forever()

    def get_on_open(self, config):
        # the methods passed to websocketClient have to be unbounded if we want WebSocketApp to pass everything correctly
        # see websocket-client/#471
        def on_open(ws):
            ws.send(json.dumps(config))

        return on_open

    def get_on_msg_cb(self):
        def on_msg_cb_local(ws, msg):
            self.on_msg_cb(msg)
            return

        return on_msg_cb_local

    # since methods need to be unbound in order for websocketClient these methods don't have a self as their first parameter
    def on_error(ws, error):  # lgtm[py/not-named-self]
        print(error)

    def on_close(ws):  # lgtm[py/not-named-self]
        pass



# -*- coding: utf-8 -*-

from collections import namedtuple
import csv
import json
import os
import re
import sys
import pkg_resources
from zipfile import ZipFile

import requests

from tiingo.restclient import RestClient
from tiingo.exceptions import (
    InstallPandasException,
    APIColumnNameError,
    InvalidFrequencyError,
    MissingRequiredArgumentError,
)

try:
    import pandas as pd

    pandas_is_installed = True
except ImportError:
    pandas_is_installed = False

VERSION = pkg_resources.get_distribution("tiingo").version


# These methods enable python 2 + 3 compatibility.
def get_zipfile_from_response(response):
    if sys.version_info < (3, 0):  # python 2
        from StringIO import StringIO as Buffer
    else:  # python 3
        from io import BytesIO as Buffer
    buffered = Buffer(response.content)
    return ZipFile(buffered)


def get_buffer_from_zipfile(zipfile, filename):
    if sys.version_info < (3, 0):  # python 2
        from StringIO import StringIO

        return StringIO(zipfile.read(filename))
    else:  # python 3
        # Source:
        # https://stackoverflow.com/questions/5627954/py3k-how-do-you-read-a-file-inside-a-zip-file-as-text-not-bytes
        from io import TextIOWrapper, BytesIO

        return TextIOWrapper(BytesIO(zipfile.read(filename)))


def dict_to_object(item, object_name):
    """Converts a python dict to a namedtuple, saving memory."""
    fields = item.keys()
    values = item.values()
    return json.loads(
        json.dumps(item), object_hook=lambda d: namedtuple(object_name, fields)(*values)
    )


class TiingoClient(RestClient):
    """Class for managing interactions with the Tiingo REST API

    Supply API Key via Environment Variable TIINGO_API_KEY
    or via the Config Object
    """

    def __init__(self, *args, **kwargs):
        super(TiingoClient, self).__init__(*args, **kwargs)
        self._base_url = "https://api.tiingo.com"

        try:
            api_key = self._config["api_key"]
        except KeyError:
            api_key = os.environ.get("TIINGO_API_KEY")
        self._api_key = api_key

        if not (api_key):
            raise RuntimeError(
                "Tiingo API Key not provided. Please provide"
                " via environment variable or config argument."
            )

        self._headers = {
            "Authorization": "Token {}".format(api_key),
            "Content-Type": "application/json",
            "User-Agent": "tiingo-python-client {}".format(VERSION),
        }

        self._frequency_pattern = re.compile("^[0-9]+(min|hour)$", re.IGNORECASE)

    def __repr__(self):
        return '<TiingoClient(url="{}")>'.format(self._base_url)

    def _is_eod_frequency(self, frequency):
        return frequency.lower() in ["daily", "weekly", "monthly", "annually"]

    # TICKER PRICE ENDPOINTS
    # https://api.tiingo.com/docs/tiingo/daily
    def list_tickers(self, assetTypes=[]):
        """Return a list of dicts of metadata tickers for all supported tickers
         of the specified asset type, as well as metadata about each ticker.
         This includes supported date range, the exchange the ticker is traded
         on, and the currency the stock is traded on.
        Tickers for unrelated products are omitted.
        https://apimedia.tiingo.com/docs/tiingo/daily/supported_tickers.zip
        """
        listing_file_url = (
            "https://apimedia.tiingo.com/docs/tiingo/daily/supported_tickers.zip"
        )
        response = requests.get(listing_file_url)
        zipdata = get_zipfile_from_response(response)
        raw_csv = get_buffer_from_zipfile(zipdata, "supported_tickers.csv")
        reader = csv.DictReader(raw_csv)

        if not len(assetTypes):
            return [row for row in reader]

        assetTypesSet = set(assetTypes)
        return [row for row in reader if row.get("assetType") in assetTypesSet]

    def list_stock_tickers(self):
        return self.list_tickers(["Stock"])

    def list_etf_tickers(self):
        return self.list_tickers(["ETF"])

    def list_fund_tickers(self):
        return self.list_tickers(["Mutual Fund"])

    def get_ticker_metadata(self, ticker, fmt="json"):
        """Return metadata for 1 ticker
        Use TiingoClient.list_tickers() to get available options

         Args:
             ticker (str) : Unique identifier for stock
        """
        url = "tiingo/daily/{}".format(ticker)
        response = self._request("GET", url)
        data = response.json()
        if fmt == "json":
            return data
        elif fmt == "object":
            return dict_to_object(data, "Ticker")

    def _invalid_frequency(self, frequency):
        """
        Check to see that frequency was specified correctly
        :param frequency (string): frequency string
        :return (boolean):
        """
        is_valid = self._is_eod_frequency(frequency) or re.match(
            self._frequency_pattern, frequency
        )
        return not is_valid

    def _get_url(self, ticker, frequency):
        """
        Return url based on frequency.  Daily, weekly, or yearly use Tiingo
        EOD api; anything less than daily uses the iex intraday api.
        :param ticker (string): ticker to be embedded in the url
        :param frequency (string): valid frequency per Tiingo api
        :return (string): url
        """
        if self._invalid_frequency(frequency):
            etext = (
                "Error: {} is an invalid frequency.  Check Tiingo API documentation "
                "for valid EOD or intraday frequency format."
            )
            raise InvalidFrequencyError(etext.format(frequency))
        else:
            if self._is_eod_frequency(frequency):
                return "tiingo/daily/{}/prices".format(ticker)
            else:
                return "iex/{}/prices".format(ticker)

    def _request_pandas(self, ticker, metric_name, params):
        """
        Return data for ticker as a pandas.DataFrame if metric_name is not
        specified or as a pandas.Series if metric_name is specified.

        :param ticker (string): ticker to be requested
        :param params (dict): a dictionary containing valid resampleFreq
            and format strings per the Tiingo api
        :param metric_name (string): Optional parameter specifying metric to be returned for each
            ticker.  In the event of a single ticker, this is optional and if not specified
            all of the available data will be returned.  In the event of a list of tickers,
            this parameter is required.
        """
        url = self._get_url(ticker, params["resampleFreq"])
        response = self._request("GET", url, params=params)
        if params["format"] == "csv":
            if sys.version_info < (3, 0):  # python 2
                from StringIO import StringIO
            else:  # python 3
                from io import StringIO

            df = pd.read_csv(StringIO(response.content.decode("utf-8")))
        else:
            df = pd.DataFrame(response.json())

        df.set_index("date", inplace=True)

        if metric_name is not None:
            prices = df[metric_name]
        else:
            prices = df

        prices.index = pd.to_datetime(prices.index)

        # Localize to UTC to ensure equivalence between data returned in json format and
        # csv format. Tiingo daily data requested in csv format does not include a timezone.
        if prices.index.tz is None:
            prices.index = prices.index.tz_localize("UTC")

        return prices

    def get_ticker_price(
        self, ticker, startDate=None, endDate=None, fmt="json", frequency="daily"
    ):
        """By default, return latest EOD Composite Price for a stock ticker.
        On average, each feed contains 3 data sources.

         Supported tickers + Available Day Ranges are here:
         https://apimedia.tiingo.com/docs/tiingo/daily/supported_tickers.zip

         Args:
             ticker (string): Unique identifier for stock ticker
             startDate (string): Start of ticker range in YYYY-MM-DD format
             endDate (string): End of ticker range in YYYY-MM-DD format
             fmt (string): 'csv' or 'json'
             frequency (string): Resample frequency
        """
        url = self._get_url(ticker, frequency)
        params = {
            "format": fmt if fmt != "object" else "json",  # conversion local
            "resampleFreq": frequency,
        }

        if startDate:
            params["startDate"] = startDate
        if endDate:
            params["endDate"] = endDate

        # TODO: evaluate whether to stream CSV to cache on disk, or
        # load as array in memory, or just pass plain text
        response = self._request("GET", url, params=params)
        if fmt == "json":
            return response.json()
        elif fmt == "object":
            data = response.json()
            return [dict_to_object(item, "TickerPrice") for item in data]
        else:
            return response.content.decode("utf-8")

    def get_dataframe(
        self,
        tickers,
        startDate=None,
        endDate=None,
        metric_name=None,
        frequency="daily",
        fmt="json",
    ):
        """Return a pandas.DataFrame of historical prices for one or more ticker symbols.

        By default, return latest EOD Composite Price for a list of stock tickers.
        On average, each feed contains 3 data sources.

        Supported tickers + Available Day Ranges are here:
        https://apimedia.tiingo.com/docs/tiingo/daily/supported_tickers.zip
        or from the TiingoClient.list_tickers() method.

        Args:
            tickers (string/list): One or more unique identifiers for a stock ticker.
            startDate (string): Start of ticker range in YYYY-MM-DD format.
            endDate (string): End of ticker range in YYYY-MM-DD format.
            metric_name (string): Optional parameter specifying metric to be returned for each
                ticker.  In the event of a single ticker, this is optional and if not specified
                all of the available data will be returned.  In the event of a list of tickers,
                this parameter is required.
            frequency (string): Resample frequency (defaults to daily).
            fmt (string): 'csv' or 'json'
        """

        valid_columns = {
            "open",
            "high",
            "low",
            "close",
            "volume",
            "adjOpen",
            "adjHigh",
            "adjLow",
            "adjClose",
            "adjVolume",
            "divCash",
            "splitFactor",
        }

        if metric_name is not None and metric_name not in valid_columns:
            raise APIColumnNameError("Valid data items are: " + str(valid_columns))

        if metric_name is None and isinstance(tickers, list):
            raise MissingRequiredArgumentError(
                """When tickers is provided as a list, metric_name is a required argument.
            Please provide a metric_name, or call this method with one ticker at a time."""
            )

        params = {"format": fmt, "resampleFreq": frequency}
        if startDate:
            params["startDate"] = startDate
        if endDate:
            params["endDate"] = endDate

        if pandas_is_installed:
            if type(tickers) is str:
                prices = self._request_pandas(
                    ticker=tickers, params=params, metric_name=metric_name
                )
            else:
                prices = pd.DataFrame()
                for stock in tickers:
                    ticker_series = self._request_pandas(
                        ticker=stock, params=params, metric_name=metric_name
                    )
                    ticker_series = ticker_series.rename(stock)
                    prices = pd.concat([prices, ticker_series], axis=1, sort=True)

            return prices

        else:
            error_message = (
                "Pandas is not installed, but .get_ticker_price() was "
                "called with fmt=pandas.  In order to install tiingo with "
                "pandas, reinstall with pandas as an optional dependency. \n"
                "Install tiingo with pandas dependency: 'pip install tiingo[pandas]'\n"
                "Alternatively, just install pandas: pip install pandas."
            )
            raise InstallPandasException(error_message)

    # NEWS FEEDS
    # tiingo/news
    def get_news(
        self,
        tickers=[],
        tags=[],
        sources=[],
        startDate=None,
        endDate=None,
        limit=100,
        offset=0,
        sortBy="publishedDate",
        onlyWithTickers=False,
        fmt="json",
    ):
        """Return list of news articles matching given search terms
        https://api.tiingo.com/docs/tiingo/news

        # Dates are in YYYY-MM-DD Format.

        Args:
            tickers [string] : List of unique Stock Tickers to search
            tags [string] : List of topics tagged by Tiingo Algorithms
            sources [string]: List of base urls to include as news sources
            startDate, endDate [date]: Boundaries of news search window
            limit (int): Max results returned. Default 100, max 1000
            offset (int): Search results offset, used for paginating
            sortBy (string): "publishedDate" OR "crawlDate", descending
            onlyWithTickers (bool): If true, only links with tagged tickers will return.
        """
        url = "tiingo/news"
        params = {
            "limit": limit,
            "offset": offset,
            "sortBy": sortBy,
            "tickers": tickers,
            "source": (",").join(sources) if sources else None,
            "tags": tags,
            "startDate": startDate,
            "endDate": endDate,
            "onlyWithTickers": onlyWithTickers,
        }
        response = self._request("GET", url, params=params)
        data = response.json()
        if fmt == "json":
            return data
        elif fmt == "object":
            return [dict_to_object(item, "NewsArticle") for item in data]

    def get_bulk_news(self, file_id=None, fmt="json"):
        """Only available to institutional clients.
        If ID is NOT provided, return array of available file_ids.
        If ID is provided, provides URL which you can use to download your
        file, as well as some metadata about that file.
        """
        if file_id:
            url = "tiingo/news/bulk_download/{}".format(file_id)
        else:
            url = "tiingo/news/bulk_download"

        response = self._request("GET", url)
        data = response.json()
        if fmt == "json":
            return data
        elif fmt == "object":
            return dict_to_object(data, "BulkNews")

    # Crypto
    # tiingo/crypto
    def get_crypto_top_of_book(
        self,
        tickers=[],
        exchanges=[],
        includeRawExchangeData=False,
        convertCurrency=None,
    ):
        url = "tiingo/crypto/top"
        params = {"tickers": ",".join(tickers)}

        if len(exchanges):
            params["exchanges"] = ",".join(exchanges)
        if includeRawExchangeData is True:
            params["includeRawExchangeData"] = True
        if convertCurrency:
            params["convertCurrency"] = convertCurrency

        response = self._request("GET", url, params=params)
        return response.json()

    def get_crypto_price_history(
        self,
        tickers=[],
        baseCurrency=None,
        startDate=None,
        endDate=None,
        exchanges=[],
        consolidateBaseCurrency=False,
        includeRawExchangeData=False,
        resampleFreq=None,
        convertCurrency=None,
    ):
        url = "tiingo/crypto/prices"
        params = {"tickers": ",".join(tickers)}

        if startDate:
            params["startDate"] = startDate
        if endDate:
            params["endDate"] = endDate
        if len(exchanges):
            params["exchanges"] = ",".join(exchanges)
        if consolidateBaseCurrency is True:
            params["consolidateBaseCurrency"] = ",".join(consolidateBaseCurrency)
        if includeRawExchangeData is True:
            params["includeRawExchangeData"] = includeRawExchangeData
        if resampleFreq:
            params["resampleFreq"] = resampleFreq
        if convertCurrency:
            params["convertCurrency"] = convertCurrency

        response = self._request("GET", url, params=params)
        return response.json()

    def get_crypto_metadata(self, tickers=[], fmt="json"):
        url = "tiingo/crypto"

        params = {
            "tickers": ",".join(tickers),
            "format": fmt,
        }

        response = self._request("GET", url, params=params)
        if fmt == "csv":
            return response.content.decode("utf-8")
        else:
            return response.json()

    # FUNDAMENTAL DEFINITIONS
    # tiingo/fundamentals/definitions
    def get_fundamentals_definitions(self, tickers=[], fmt="json"):
        """Return definitions for fundamentals for specified tickers
        https://api.tiingo.com/documentation/fundamentals

        Args:
            tickers [string] : optional, either list or string
            fmt (string): 'csv' or 'json'
        """
        url = "tiingo/fundamentals/definitions"
        params = {"tickers": tickers, "format": fmt}
        response = self._request("GET", url, params=params)
        if fmt == "json":
            return response.json()
        elif fmt == "csv":
            return response.content.decode("utf-8")

    # FUNDAMENTAL DAILY
    # tiingo/fundamentals/<ticker>/daily
    def get_fundamentals_daily(self, ticker, fmt="json", startDate=None, endDate=None):
        """Returns metrics which rely on daily price-updates
        https://api.tiingo.com/documentation/fundamentals

        # Dates are in YYYY-MM-DD Format.

        Args:
            tickers [string] : List of unique Stock Tickers to search
            startDate, endDate [date]: Boundaries of search window
            fmt (string): 'csv' or 'json'
        """
        url = "tiingo/fundamentals/{}/daily".format(ticker)
        params = {"startDate": startDate, "endDate": endDate, "format": fmt}
        response = self._request("GET", url, params=params)
        if fmt == "json":
            return response.json()
        elif fmt == "csv":
            return response.content.decode("utf-8")

    # FUNDAMENTAL STATEMENTS
    # tiingo/fundamentals/<ticker>/statements
    def get_fundamentals_statements(
        self, ticker, asReported=False, fmt="json", startDate=None, endDate=None
    ):
        """Returns data that is extracted from quarterly and annual statements.
            https://api.tiingo.com/documentation/fundamentals

            # Dates are in YYYY-MM-DD Format.

            Args:
                tickers [string] : List of unique Stock Tickers to search
                startDate, endDate [date]: Boundaries of search window
                asReported [bool]: get most-recent data (False) or data \
                                   as it was reported on the release-date
                                   (True)
                fmt (string): 'csv' or 'json'
        """
        if asReported:
            asReported = "true"
        else:
            asReported = "false"

        url = "tiingo/fundamentals/{}/statements".format(ticker)
        params = {
            "startDate": startDate,
            "endDate": endDate,
            "asReported": asReported,
            "format": fmt,
        }
        response = self._request("GET", url, params=params)
        if fmt == "json":
            return response.json()
        elif fmt == "csv":
            return response.content.decode("utf-8")



#!/usr/bin/env python
"""Tests for `tiingo` package."""

from unittest import TestCase

import vcr

from tiingo import TiingoClient


# CRYPTO ENDPOINTS
class TestCryptoEndpoints(TestCase):

    def setUp(self):
        self._client = TiingoClient()

    @vcr.use_cassette('tests/fixtures/crypto_metadata.yaml')
    def test_crypto_metadata(self):
        metadata = self._client.get_crypto_metadata(tickers=['btcusd', 'fldcbtc'])

        assert len(metadata) == 2 # 1 item per ticker
        assert metadata[0]['ticker'] == 'btcusd'
        assert metadata[1]['ticker'] == 'fldcbtc'

    @vcr.use_cassette('tests/fixtures/crypto_top_of_book.yaml')
    def test_crypto_top_of_book(self):
        top_of_book = self._client.get_crypto_top_of_book(tickers=['btcusd', 'fldcbtc'], includeRawExchangeData=True)
        assert len(top_of_book) == 2 # 1 each for btcusd and fldcbtc
        first_ticker_data = top_of_book[0]
        assert 'topOfBookData' in first_ticker_data
        assert 'exchangeData' in first_ticker_data

    @vcr.use_cassette('tests/fixtures/crypto_price_history.yaml')
    def test_crypto_price_history(self):
        price_history = self._client.get_crypto_price_history(tickers=['btcusd', 'fldcbtc'],
                                                              startDate='2019-05-01', endDate='2019-05-02',
                                                              includeRawExchangeData=True, resampleFreq='1day')
        assert len(price_history) == 2 # 1 item per ticker
        assert len(price_history[0]['priceData']) == 2 # 1 set of data per day
        assert 'exchangeData' in price_history[0] # includedRawExchangeData



#!/usr/bin/env python
"""Unit tests for pandas functionality in tiingo"""

import vcr
from unittest import TestCase
from tiingo import TiingoClient
from tiingo.exceptions import APIColumnNameError, InstallPandasException, MissingRequiredArgumentError
try:
    import pandas as pd
    pandas_is_installed = True
except ImportError:
    pandas_is_installed = False


class TestTiingoWithPython(TestCase):

    def setUp(self):
        if pandas_is_installed:
            self._client = TiingoClient()
        else:
            self.skipTest("test_tiingo_pandas: Pandas not installed.")

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_weekly.yaml')
    def test_return_pandas_format(self):
        """Test that valid pandas format is returned when specified"""
        prices = self._client.get_dataframe("GOOGL", startDate='2018-01-05',
                                            endDate='2018-01-19', frequency='weekly')
        self.assertTrue(isinstance(prices, pd.DataFrame))
        assert len(prices.index) == 3

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_weekly_multiple_tickers.yaml')
    def test_return_pandas_format_multiple(self):
        """Test that valid pandas format is returned when specified"""
        tickers = ["GOOGL", "AAPL"]
        prices = self._client.get_dataframe(tickers, startDate='2018-01-05',
                                            endDate='2018-01-19', metric_name='adjClose', frequency='weekly')
        self.assertTrue(isinstance(prices, pd.DataFrame))
        assert len(prices.columns) == 2
        assert len(prices.index) == 3

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_weekly_multiple_tickers_csv.yaml')
    def test_return_pandas_format_multiple(self):
        """Test that valid pandas format is returned when specified"""
        tickers = ["GOOGL", "AAPL"]
        prices = self._client.get_dataframe(tickers, startDate='2018-01-05',
            endDate='2018-01-19', metric_name='adjClose', frequency='weekly', fmt='csv')
        self.assertTrue(isinstance(prices, pd.DataFrame))
        assert len(prices.columns) == 2
        assert len(prices.index) == 3

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_daily.yaml')
    def test_return_pandas_daily(self):
        """Test that valid pandas format is returned when specified"""
        prices = self._client.get_dataframe("GOOGL", startDate='2018-01-05',
                                            endDate='2018-01-19', frequency='daily')
        self.assertTrue(isinstance(prices, pd.DataFrame))
        assert len(prices.columns) == 12

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_daily_csv.yaml')
    def test_return_pandas_daily_csv(self):
        """Test that valid pandas format is returned when specified and csv data requested"""
        prices = self._client.get_dataframe("GOOGL",
                                            startDate='2018-01-05', endDate='2018-01-19',
                                            frequency='daily', fmt='csv')
        self.assertTrue(isinstance(prices, pd.DataFrame))
        self.assertTrue(isinstance(prices.index, pd.DatetimeIndex))
        assert len(prices) == 10
        assert len(prices.columns) == 12

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_daily_metric_name.yaml')
    def test_return_pandas_daily(self):
        """Test that one column is returned when a metric name is specified"""

        prices = self._client.get_dataframe("GOOGL", startDate='2018-01-05', metric_name='adjClose',
                                            endDate='2018-01-19', frequency='daily')
        self.assertTrue(isinstance(prices, pd.Series))
        assert len(prices.index) == 10

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_daily_metric_name_csv.yaml')
    def test_return_pandas_daily_metric_name_csv(self):
        """Test that one column is returned when a metric name is specified and csv data requested
           Request unadjusted close column to ensure data remains constant in case GOOGL splits
           or distributes dividends.
        """
        prices = self._client.get_dataframe("GOOGL", startDate='2018-01-05', metric_name='close',
                                            endDate='2018-01-19', frequency='daily', fmt='csv')
        self.assertTrue(isinstance(prices, pd.Series))
        self.assertTrue(isinstance(prices.index, pd.DatetimeIndex))
        assert prices.values.tolist() == [
            1110.29,1114.21,1112.79,1110.14,1112.05,
            1130.65,1130.7,1139.1,1135.97,1143.5]
        assert len(prices.index) == 10

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_daily_equivalent_requesting_json_or_csv.yaml')
    def test_price_pandas_daily_equivalent_requesting_json_or_csv(self):
        """Test that equivalent data is returned when specifying reuqest format in json or csv.
        """
        prices_json = self._client.get_dataframe("GOOGL",
            startDate='2018-01-05', endDate='2018-01-19',
            metric_name='close', frequency='daily')

        prices_csv = self._client.get_dataframe("GOOGL",
            startDate='2018-01-05', endDate='2018-01-19',
            metric_name='close', frequency='daily', fmt='csv')

        self.assertTrue(prices_json.equals(prices_csv))

    @vcr.use_cassette('tests/fixtures/intraday_price.yaml')
    def test_intraday_ticker_price(self):
        """Test the EOD Prices Endpoint with data param"""
        prices = self._client.get_dataframe("GOOGL",
                                            startDate="2018-01-02",
                                            endDate="2018-01-02",
                                            frequency="30Min")
        self.assertGreater(len(prices), 1)

    def test_metric_name_column_error(self):
        with self.assertRaises(APIColumnNameError):
            self._client.get_dataframe(['GOOGL', 'AAPL'], startDate='2018-01-05',
                                                endDate='2018-01-19', metric_name='xopen', frequency='weekly')

    def test_metric_name_missing_when_multiple_tickers(self):
        with self.assertRaises(MissingRequiredArgumentError):
            self._client.get_dataframe(['GOOGL', 'AAPL'], frequency='weekly')

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_single.yaml')
    def test_pandas_edge_case(self):
        """Test single price/date being returned as a frame"""
        prices = self._client.get_dataframe("GOOGL")
        assert len(prices) == 1
        assert len(prices.index) == 1


class TestTiingoWithoutPython(TestCase):

    def setUp(self):
        if pandas_is_installed:
            self.skipTest("test_tiingo_without_pandas: Pandas is installed.")
        else:
            self._client = TiingoClient()

    @vcr.use_cassette('tests/fixtures/ticker_price_pandas_single.yaml')
    def test_get_dataframe_without_pandas(self):
        with self.assertRaises(InstallPandasException):
            self._client.get_dataframe("GOOGL")



# Exception Clasess
class InstallPandasException(Exception):
    pass


class APIColumnNameError(Exception):
    pass


class InvalidFrequencyError(Exception):
    pass


class MissingRequiredArgumentError(Exception):
    pass



import os
from unittest import TestCase,mock
from tiingo.wsclient import TiingoWebsocketClient
from tiingo.exceptions import MissingRequiredArgumentError

class TestRestClientWithSession(TestCase):
    def setUp(self):

        def msg_cb(msg):
            print(msg)

        self.cb=msg_cb

        self.config =  {
            'eventName':'subscribe',
            'authorization':os.getenv("TIINGO_API_KEY"),
            #see https://api.tiingo.com/documentation/websockets/iex > Request for more info
            'eventData': { 
                'thresholdLevel':5
            }
        }

    # test for missing or incorrectly supplied endpoints
    def test_missing_or_wrong_endpoint(self):
        with self.assertRaises(AttributeError) as ex:
            TiingoWebsocketClient(config=self.config,on_msg_cb=self.cb)
        self.assertTrue(type(ex.exception)==AttributeError)

        with self.assertRaises(AttributeError) as ex:
             TiingoWebsocketClient(config=self.config,endpoint='wq',on_msg_cb=self.cb)
        self.assertTrue(type(ex.exception)==AttributeError)

    # test for missing callback argument    
    def test_missing_msg_cb(self):
        with self.assertRaises(MissingRequiredArgumentError) as ex:
            TiingoWebsocketClient(config=self.config,endpoint='iex')
        self.assertTrue(type(ex.exception)==MissingRequiredArgumentError)

    # test for missing API keys in config dict and in os env
    def test_missing_api_key(self):
        with mock.patch.dict(os.environ, {}, clear=True): #clear env vars including the TIINGO_API_KEY
            with self.assertRaises(RuntimeError) as ex:
                TiingoWebsocketClient(config={},endpoint='iex',on_msg_cb=self.cb)
            self.assertTrue(type(ex.exception)==RuntimeError)
